{==============================================================================
      Unit:  TypeMan.pas
    Status:  Freeware - No restrictions apply - without a warranty of any kind.
  Compiler:  Delphi4, Delphi5, Delphi6
    Author:  Eric Grobler (egrobler@quoka.com)
   Version:  1.21
       Url:  www.geocities.com/ericdelphi/TypeMan.html
  Lastdate:  28 Sep 2001
   Purpose:  Classes to manage published properties and functions that extends
             the functionality of the rtti routines in TypInfo.pas

History:
  26 Aug 2001 - First version
  10 Sep 2001 - Changed overloaded constructers
  21 Sep 2001 - ClassInfo check to allow $M+ classes. (contribution by Martijn Brinkers)
  28 Sep 2001 - Added IsOfClass, GetEnumValues, AssignDefaultValue, ToDataSet, FromDataSet etc.
==============================================================================}
unit TypeMan;

interface

uses Classes, SysUtils, contnrs, Consts, Controls, Db,
     {$IfDef VER140} //Delphi6
       Variants,
     {$EndIf}
      TypInfo;

type
{-------------------------------------------------------------------------------
TTypeUtils
  A collection of TypeInfo related functions
  If TMyColor = (Red, Green, Blue) then:
    Example1:
      EnumToString(TypeInfo(TMyColor), Byte(MyColorVar)) > 'Green'
    Example2:
      TypeUtils.EnumValuesToString(TypeInfo(TMyColor)) will return 'Red','Green','Blue'
    Example3:
      TypeUtils.SetProperties(Form1, 'Color', clYellow, TypeInfo(TMyColor))
      will set all visual control's to yellow!
--------------------------------------------------------------------------------

{-------------------------------------------------------------------------------
TQPropList
  A list to manage any published fields in a Delphi (normally TPersistent decendent) class.
  Note: see Delphi Help on the $M+ directive and TPersistent.
  
  QPropList reads the type info (via rtti) of published fields into an internal list.
  You can either assign a "ClassType" or a "ClassInstance" to the list:
  This will load the "TypeInfo" of the published fields of the TForm ClassType:
    PropList.ForType := TForm;
  This will also load the addresses for read/write access of this particular instance:
    PropList.ForInstance := Form1;
  Once a class is assigned, the list provides easy access to the fields.
    PropList.ByName('Color').AsString := 'clGreen';

  You can either assign a class in the constructor, or by repeatedly setting new classes
  you wish to view/manipulate via properties ForType/ForInstance.

  Usage Examples:
    PropList.ForInstance := Form1; (load's the publsihed fields)
    if PropList.PropExists('Anchors') then
       PropList.ByName('Anchors').AsString := 'akLeft,akTop';
    if PropList.PropExists('Font')
     then if PropList.ByName('Font').TypeInfoPtr=TypeInfo(TFont)
      then (PropList.ByName('Font').AsClass as TFont).Style := [fsBold];
    for i := 1 to PropList.Count-1 do Memo1.Lines.Add(PropList.Items[i].Name);


  Small Demo:
    TPerson = class(TPersistent)
    private
     FSurName : String;
     FAge : Integer;
     FColor : TMyColor;
    published
     property SurName : String read FSurName write FSurName;
     property Age : Integer read FAge write FAge;
     propery Color : TMyColor read FColor write FColor;
    end;

    procedure TForm1.Test;
    var aPerson : TPerson;
     aPropList : TQPropList;
    begin
     aPerson := TPerson.Create;
     aPropList := TQPropList.Create(aPerson);
     aPropList.ByName('SurName').AsString := 'Borland';
     aPropList.ByName('Age').AsInteger := 22;
     aPropList.ByName('Color').AsString := 'clGreen';
     (you can also use the default array definition)
     aPropList[Color'].AsString := 'clGreen';
     aPropList.ValuesToSL(Memo1.Lines, True);
     aPropList.Free;
     aPerson.Free;
    end;


Notes on "DirtyWrite"
  Consider the Age property declaration:
    TPerson = class(TPersistent)
    protected
      FAge    : Byte;
      FSalary : Currency;
    published
      property Age : Byte read FAge;
      property Salary : Currency read GetCurrency;
    end;
  Although a "write specifier" was not provided we can safely use the "read address"
  to write to the Age property.
  However the Salary propery makes use of a method therefore writing is not possible.
  If you use DirtyWrite in TQPropList and if a property does not have a "write specifier",
  it will use the GetProc address for writing, if the address is a pointer to a
  variable and not a method.

  In addition you can also use properties DirectPtr and IsDirectAccess (without DirtyWrite set)
    var p : ^Byte;
    begin
      if not PropList.ByName('Age').IsDirectAccess then Exit;
      p := PropList.ByName('Age').DirectPtr;
      p^ := 77;
      - or -
      Byte(PropList.ByName('Age').DirectPtr^) := 77;


-------------------------------------------------------------------------------}

  TQPropTypeAccess = (taNone, taInt, taInt64, taChar, taBoolean, taString,
                      taDateTime, taFloat, taEnum, taSet, taVariant, taClass);

  {Primitive Delphi types - see also mf.PrimTypeByteSize}
  TPrimType = (
    ptNone,
    ptByte,         // 1     ord
    ptShortInt,     // 1     ord
    ptBoolean,      // 1     ord
    ptChar,         // 1     ord
    ptWord,         // 2     ord
    ptSmallInt,     // 2     ord
    ptWideChar,     // 2     ord
    ptInteger,      // 4     ord
    ptLongint,      // 4     ord
    ptLongword,     // 4     ord
    ptCardinal,     // 4     ord  (identical to Longword)
    ptPointer,      // 4     pointer
    ptClass,        // 4     pointer
    ptSingle,       // 4     Real
//  ptReal48,       // 6     Real           legacy
    ptReal,         // 8     Real           same as double?
    ptInt64,        // 8     ord
    ptDouble,       // 8     Real
    ptComp,         // 8     Real
    ptCurrency,     // 8     Real
    ptTDateTime,    // 8     Real
    ptTDate,        // 8     Real
    ptExtended,     //10     Real
    ptAnsiString,   // 4     String  2GB    pointer
    ptWideString,   // 4     String  2GB    pointer (2byte chars)
    ptEnum,         //-1     ord     1..4   (Normally 1 Byte(256 elements)
    ptSet,          //-1     ord     1..16  (Only supports 1 Byte enums)
    ptVariant,      //16
    ptShortString); //-1     String  1..256 Byte array
  TPrimTypes = set of TPrimType;
const
  PrimTypeIntSet = [ptByte, ptShortInt, ptWord, ptSmallInt, ptInteger, ptLongint, ptLongword, ptCardinal, ptInt64];
  PrimTypeOrdSet = PrimTypeIntSet + [ptBoolean, ptChar, ptWideChar, ptEnum, ptSet];
  PrimTypeFloatSet = [ptSingle, ptReal, ptDouble, ptExtended, ptComp, ptCurrency];
  PrimTypeDateSet = [ptTDateTime, ptTDate];
  PrimTypeStrSet = [ptAnsiString, ptWideString, ptShortString];

type
  TOpenStringArray = array of String;

  PObjectOfProp = ^TObjectOfProp;
  TObjectOfProp = record
    FType       : TClass;
    FInstance   : TObject;
    FDirtyWrite : Boolean;
  end;

const
  {TTypeKinds supported}
  TypeManSupportedTypeKinds : TTypeKinds = [tkInteger, tkChar, tkEnumeration, tkFloat, tkString, tkSet, tkInt64, tkLString, tkVariant, tkClass];

type
  TQPropTypesAccess = set of TQPropTypeAccess;
  TQCustomPropItem = class
  private
    function GetPropClassType: TClass;
  protected
    FObjectPtr   : PObjectOfProp;
    FPropInfo    : TPropInfo;
    FTypeData    : PTypeData;
    FPrimType    : TPrimType;
    FTypeAccess  : TQPropTypeAccess;
    FVarByteSize : Byte;

    procedure DoRaise(const aMsg : String;const aFunc : String = '');
    procedure ClearInfoValues; virtual;
    procedure SetDirtyWrite;
    function  GetLongestEnumValue : Integer;
    function  GetLongestSetValue : Integer;
    function  GetEnumValues : String;


    function  DeterminePrimitiveType : TPrimType;
    function  DetermineDirectPtr : Pointer;
    function  DetermineTypeAccess : TQPropTypeAccess;
    function  DetermineVarByteSize : Byte; //Classes and AnsiStrings will be SizeOf(Pointer)
    function  GetName: String;
    function  GetTypeKind: TTypeKind;
    function  GetDefaultValue : Integer;
    function  GetHasDefaultValue : Boolean;

    function  GetAsString: String;
    procedure SetAsString(const Value: String);
    function  GetAsInteger: Integer;
    procedure SetAsInteger(const Value: Integer);
    function  GetAsCardinal: Cardinal;
    procedure SetAsCardinal(const Value: Cardinal);
    function  GetAsChar: Char;
    procedure SetAsChar(const Value: Char);
    function  GetAsBoolean: Boolean;
    procedure SetAsBoolean(const Value: Boolean);
    function  GetAsFloat: Extended;
    procedure SetAsFloat(const Value: Extended);
    function  GetAsDateTime: TDateTime;
    procedure SetAsDateTime(const Value: TDateTime);
    function  GetAsVariant: Variant;
    procedure SetAsVariant(const Value: Variant);
    function GetPrimTypeAsStr: String;
    function GetTypeName: String;
    function GetAsInt64: Int64;
    procedure SetAsInt64(const Value: Int64);
    function GetHasMinMaxValue : Boolean;
    function GetMaxValue: Int64;
    function GetMinValue : Int64;
    function GetTypeInfoPtr: PTypeInfo;

    function DataOk: Boolean;
    procedure CheckWriteOk;
    procedure CheckReadOk;
    function GetIsDirectAccess: Boolean;
    function GetDirectPtr: Pointer;
    function GetAsClass: TObject;
    procedure SetAsClass(const Value: TObject);
    function GetAsPointer: Pointer;
    procedure SetAsPointer(const Value: Pointer);
    function GetReadOnly: Boolean;
    procedure SetForInstance(const Value: TObject);
    function GetForType: TClass;
    function GetForInstance: TObject;
    function GetDirtyWrite: Boolean;
    function GetClassInUnit : String;

    procedure DoSetProperty(ForAObject : PObjectOfProp; aPropInfo : PPropInfo);
    function GetPropInfoPtr: PPropInfo;
    constructor Create(ForAObject : PObjectOfProp; aPropInfo : PPropInfo); overload;
  public
    property ForInstance : TObject read GetForInstance write SetForInstance;
    property ForType : TClass read GetForType;

    //Info
    property Name : String read GetName;
    property TypeName : String read GetTypeName;

    property MaxValue : Int64 read GetMaxValue; //Enums are 0 based
    property MinValue : Int64 read GetMinValue;
    property HasMinMaxValue : Boolean read GetHasMinMaxValue;
    property VarByteSize : Byte read FVarByteSize;  //AnsiString will be SizeOf(Pointer), ShortStrings will be SizeOf

    property TypeKind : TTypeKind read GetTypeKind;
    property PrimType : TPrimType read FPrimType;
    property PrimTypeAsStr : String read GetPrimTypeAsStr;
    property PropClassType : TClass read GetPropClassType;
    function IsOfClass(aClass: TClass): Boolean; overload;
    function IsOfClass(aInstance: TObject): Boolean; overload;
    property TypeAccess : TQPropTypeAccess read FTypeAccess;

    property IsDirectAccess : Boolean read GetIsDirectAccess;
    property DefaultValue : Integer read GetDefaultValue;
    property HasDefaultValue : Boolean read GetHasDefaultValue;
    property ReadOnly : Boolean read GetReadOnly;

    property PropInfo : TPropInfo read FPropInfo;
    property PropInfoPtr : PPropInfo read GetPropInfoPtr;
    property TypeInfoPtr : PTypeInfo read GetTypeInfoPtr;
    property TypeData : PTypeData read FTypeData;
    property ClassInUnit : String read GetClassInUnit; //if property a class then UnitName where declared

    //Access
    property AsString : String read GetAsString write SetAsString;
    property AsInteger : Integer read GetAsInteger write SetAsInteger;
    property AsCardinal : Cardinal read GetAsCardinal write SetAsCardinal;
    property AsInt64 : Int64 read GetAsInt64 write SetAsInt64;
    property AsChar : Char read GetAsChar write SetAsChar;
    property AsBoolean : Boolean read GetAsBoolean write SetAsBoolean;
    property AsFloat : Extended read GetAsFloat write SetAsFloat;
    property AsDateTime : TDateTime read GetAsDateTime write SetAsDateTime;
    property AsVariant : Variant read GetAsVariant write SetAsVariant; 
    property AsClass   : TObject read GetAsClass write SetAsClass;
    property AsPointer : Pointer read GetAsPointer write SetAsPointer;

    //Special
    procedure ClearValue; virtual;
    procedure AssignValue(aSource : TQCustomPropItem);
    procedure AssignDefaultValue;
    procedure WriteToStream(aStream : TStream);
    procedure ReadFromStream(aStream : TStream);
    property  LongestEnumValue : Integer read GetLongestEnumValue;
    property  LongestSetValue : Integer read GetLongestSetValue;
    property  EnumValues  : String read GetEnumValues;
    property  DirectPtr : Pointer read GetDirectPtr;

    procedure AssignFromTField(aSource : TField);
    procedure AssignToTField(aTarget : TField);
  end;

  TQListPropItem = class(TQCustomPropItem); //for use by TQPropList

  TQPropItem = class(TQCustomPropItem) //for standalone usage
  protected
    FObjectOfProp : TObjectOfProp;
  public
    //see notes on DirtyWrite above
    constructor Create(ForAClassType : TClass; const aPropName : String; aDirtyWrite : Boolean = False); overload;
    constructor Create(ForAInstance : TObject; const aPropName : String; aDirtyWrite : Boolean = False); overload;
    procedure SetProperty(ForAClassType : TClass; const aPropName : String; aDirtyWrite : Boolean = False); overload;
    procedure SetProperty(ForAInstance : TObject; const aPropName : String; aDirtyWrite : Boolean = False); overload;
  end;

  TQPropList = class(TPersistent)
  private
    FLimitTypes : TPrimTypes;
    FIsExcludeTypes : Boolean;
    FLimitClass : TClass;

    procedure DoRaise(const aMsg : String;const aFunc : String = '');
    function  GetPropByName(const Index: String): TQListPropItem;
    procedure AssignPropFromTField(aField: TField; RaiseIfPropMissing: Boolean);
    procedure AssignPropToDataSet(aProp : TQListPropItem; aDataSet : TDataSet; RaiseIfFieldMissing : Boolean);
  protected
    FList           : TObjectList;
    FObjectForProps : TObjectOfProp;
    function  GetItem(Index: Integer): TQListPropItem;
    function  GetCount: Integer;
    function  GetAsString(Index : Integer) : String;
    function  GetPropName(Index : Integer) : String;
    procedure SetForInstance(const Value: TObject);
    procedure SetForType(const Value: TClass);
    procedure CreateVars;
    procedure FreeVars;
    procedure AddPropertyToList(aPropInfo : PPropInfo);
    procedure DoCreatePropertyList(ForAClassType : TClass; ForAInstance : TObject); virtual;
    property  PropByName[const Index : String] : TQListPropItem read GetPropByName; default; //Usage: PropList['name'].AsString :=
    function  FindProp(const aPropName: String; const MustBeOfType : PTypeInfo= nil) : TQListPropItem;
  public
    constructor Create(aDirtyWrite : Boolean = False); overload;
    constructor Create(ForAClassType : TClass; aDirtyWrite : Boolean = False; aTypeLimit : TPrimTypes = []; aClassLimit : TClass = nil); overload;
    constructor Create(ForAInstance : TObject; aDirtyWrite : Boolean = False; aTypeLimit : TPrimTypes = []; aClassLimit : TClass = nil); overload;
    constructor CreateExclude(ForAClassType : TClass; aDirtyWrite : Boolean = False; aTypeLimit : TPrimTypes = []; aClassLimit : TClass = nil); overload;
    constructor CreateExclude(ForAInstance : TObject; aDirtyWrite : Boolean = False; aTypeLimit : TPrimTypes = []; aClassLimit : TClass = nil); overload;
    destructor Destroy; override;

    //Use ForInstance/ForType to dynamically switch between instances or different classes.
    property ForInstance : TObject read FObjectForProps.FInstance write SetForInstance;
    property ForType : TClass read FObjectForProps.FType write SetForType;

    procedure Clear;
    property Items[Index : Integer] : TQListPropItem read GetItem;
    property Count : Integer read GetCount;  //"PropCount" - number of published fields added.

    property PropName[Index : Integer] : String read GetPropName;
    property AsString[Index : Integer] : String read GetAsString;

    function ByName(const aPropName : String) : TQListPropItem; //Can also use class default like PropList['aPropName'].AsString
    function IndexByName(const aPropName: String; const MustBeOfType : PTypeInfo= nil) : Integer;
    function PropExists(const aPropName : String; const MustBeOfType : PTypeInfo= nil) : Boolean;

    //Optional - to be set before using ForInstance/ForType
    property LimitTypes : TPrimTypes read FLimitTypes write FLimitTypes; //include or exclude depending on IsExcludeTypes
    property LimitClass : TClass read FLimitClass write FLimitClass;
    property IsExcludeTypes : Boolean read FIsExcludeTypes write FIsExcludeTypes;
    property DirtyWrite : Boolean read FObjectForProps.FDirtyWrite write FObjectForProps.FDirtyWrite default False;  //see notes at top

    //Extras
    procedure TypesToSL(aList : TStrings; Ident : Byte = 0; MaxValueStrLen : Integer = $ffff); //for visual feedback
    procedure ValuesToSL(aList : TStrings; WithType : Boolean = False; Ident : Byte = 0; MaxValueStrLen : Integer = $ffff); //for visual feedback
    procedure WriteToStream(aStream : TStream; IgnoreTypes : TQPropTypesAccess = [taClass]);
    procedure ReadFromStream(aStream : TStream; IgnoreTypes : TQPropTypesAccess = [taClass]);
    procedure ClearPropValues(IgnoreTypes : TQPropTypesAccess = [taClass]);
    procedure AssignDefaultValues;    

    //Assigns common properties (by name) between 2 classes
    procedure AssignPropValues(aSourceClass : TObject; IgnoreTypes : TQPropTypesAccess = [taClass]); overload;
    procedure AssignPropValues(aPropListForSourceClass  : TQPropList; IgnoreTypes : TQPropTypesAccess = [taClass]); overload;

    //Assigns fields between TDataSet and PropList
    procedure FromDataSet(aDataSet : TDataSet; RaiseIfPropMissing : Boolean = True); overload;
    procedure FromDataSet(aDataSet : TDataSet; const FieldNames : array of String; IsInclude : Boolean = True; RaiseIfPropMissing : Boolean = True); overload;
    procedure ToDataSet(aDataSet: TDataSet; RaiseIfFieldMissing: Boolean = True); overload;
    procedure ToDataSet(aDataSet : TDataSet; const PropNames : array of String; IsInclude : Boolean = True; RaiseIfFieldMissing : Boolean = True); overload;

  end;

  TTypeUtils = class
  private
    procedure DoRaise(const aMsg : String; const aFunc : String = '');
  public
    //Convert enumumerated values to a strings (see examples at implementation)
    function  EnumToString(pInfo : PTypeInfo; aOrdValue : Integer; aOffset : Integer = 1) : String;
    function  EnumValuesToString(pInfo : PTypeInfo; aOffset : Integer = 1; aDelim : String =#13#10): String;
    function  EnumValuesToArray(pInfo: PTypeInfo; aOffset : Integer = 1): TOpenStringArray;
    function  EnumLongestName(pInfo : PTypeInfo; aOffset : Integer = 1) : Integer;

    //String to enumerated type
    function  StrToEnum(pInfo : PTypeInfo; const aElementName : String; out EnumFound : Boolean; aOffset : Integer = 1; Ignorecase : Boolean = True) : Integer; overload;
    function  StrToEnum(pInfo : PTypeInfo; const aElementName : String; aOffset : Integer = 1;Ignorecase : Boolean = True) : Integer; overload;

    //Enums in a set to String (see examples at implementation)
    function  EnumSetToString(pInfo: PTypeInfo; aOrdValue : Int64; aOffset : Integer = 1; Brackets : Boolean = False; const aDelim : String = ',') : String; overload;
    function  EnumSetToString(pInfo: PTypeInfo; aValue : Pointer; aOffset : Integer = 1; Brackets : Boolean = False; const aDelim : String = ',') : String; overload;

    function  SetProperties(aOwner : TComponent; const aPropName : string; const aValue : Variant; const OfType : PTypeInfo = nil; InClassType : TClass = nil; DoInherit : Boolean = False): Integer; overload;
    function  SetProperties(aOwner : TComponent; const aPropName : string; const aValue : Variant; InClassType : TClass; DoInherit : Boolean = False): Integer; overload;

    //Special - low level
    function  EnumSize(pInfo: PTypeInfo) : Integer;
    function  EnumValuesToInt(pInfo: PTypeInfo; aValue : Pointer) : Int64;
    function  PropInfoToTypeData(aPropInfo : PPropInfo) : PTypeData;

    //PropInfoTo
    function PropInfoToClassType(aPropInfo : PPropInfo) : TClass;
    function ClassDecendantOf(aPropInfo : PPropInfo; ForClass : TClass) : Boolean;

    //Primitive Types
    function PropInfoToPrimType(aPropInfo : PPropInfo) : TPrimType;
    function PrimTypeByteSize(aPrimType: TPrimType): Byte;
  end;

var TypeUtils : TTypeUtils = nil;

implementation

function Bit64On(const i: Int64; const TheBit: Byte): Boolean;
// Returns True if a Bit in i is on (1 Based)
// 9 in binary form is 00001001 thus BitOn(9,1) is True
begin
  Result := (i and (Int64(1) shl (TheBit-1))) <> 0;
end;

function ToBool(const s : String) : Boolean;
begin
  if s=''
    then Result := False
    else Result := s[1] in ['1','T','t','Y','y','J','j'];
end;

function IsSame(const s1,s2 : String; IgnoreCase : Boolean = False) : Boolean;
begin
  if IgnoreCase
   then Result := Uppercase(Trim(s1))=Uppercase(Trim(s2))
   else Result := Trim(s1)=Trim(s2);
end;

function RepeatChar(c : Char; Count : Integer) : String;
begin
  SetLength(Result, Count);
  FillChar(Result[1], Count, c);
end;

procedure RaiseWithPrevMsg(const aMsg, aFunc: String; aClass : TObject);
var
 e : TObject;
 s : String;
   procedure Add(const aLine : string);
   begin
     if pos(aLine, s)>0 then Exit;
     if s=''
       then s := aLine
       else s:= s+#13#10+aLine;
   end;
begin
  s := '';
  e := ExceptObject;
  if e<>nil
    then Add(Exception(e).message);
  if aMsg<>''
    then Add(aMsg);
  if aFunc<>''
    then Add('Func: '+aFunc);
  if aClass<>nil
    then Add('Class: '+aClass.ClassName);
  raise exception.Create(s);
end;

{ TTypeUtils }
procedure TTypeUtils.DoRaise(const aMsg, aFunc: String);
begin
  RaiseWithPrevMsg(aMsg, aFunc, Self);
end;

{-------------------------------------------------------------------------------
Returns all the elemnents in a enumerated type (or set) as string with CR delimiters by default.
Example:
 type
  TMyColor = (Red, Green, Blue)
  TMyColors = set of TMyColor;
  Memo1.Text := EnumValuesToString(TypeInfo(TMyColor), 1, ',') will return 'Red','Green','Blue'
  Memo1.Text := EnumValuesToString(TypeInfo(TMyColors), 1, ',') will return 'Red','Green','Blue'
--------------------------------------------------------------------------------}
function TTypeUtils.EnumValuesToString(pInfo: PTypeInfo; aOffset : Integer = 1; aDelim : String =#13#10): String;
var pData : PTypeData;
        i : Integer;
begin
  Result := '';
  if aOffset<1 then aOffset := 1;

  if (pInfo^.Kind = tkSet)
    then pInfo := GetTypeData(pInfo)^.CompType^; //If a Set - get the enum type the set is based on

  if not (pInfo^.Kind = tkEnumeration)
    then DoRaise('TypeInfo not compatible', 'EnumValuesToString');

  pData := GetTypeData(pInfo);
  for i := 0 to pData^.MaxValue do
  begin
    if i=pData^.MaxValue
      then Result := Result + Copy(GetEnumName(pInfo,i), aOffset, $FF)
      else Result := Result + Copy(GetEnumName(pInfo,i), aOffset, $FF) + aDelim;
  end;
end;

{------------------------------------------------------------------------------
 Returns a dynamic array with the elements of an enumerated type as string.
 Sets are also supported
 If type TMyColor = (red,green,blue) then;
   Memo1.Text := EnumValuesToArray(TypeInfo(TMyColor)) will return 'red','green','blue'
-------------------------------------------------------------------------------}
function TTypeUtils.EnumValuesToArray(pInfo: PTypeInfo; aOffset : Integer = 1): TOpenStringArray;
var pData : PTypeData;
        i : Integer;
begin
  Result := nil;
  if aOffset<1 then aOffset := 1;

  if (pInfo^.Kind = tkSet)
    then pInfo := GetTypeData(pInfo)^.CompType^; //If a Set - get the enum type the set is based on

  if not (pInfo^.Kind = tkEnumeration)
    then DoRaise('TypeInfo not compatible', 'EnumValuesToArray');

  pData := GetTypeData(pInfo);
  for i := 0 to pData^.MaxValue do
  begin
    SetLength(Result,Length(Result)+1);
    Result[High(Result)] := Copy(GetEnumName(pInfo,i),aOffset,$FF);
  end;
end;


{------------------------------------------------------------------------------
Returns the the longest name of the elements in a enumerated type.
Works also with sets.
Example:
   TMyColor = (Red, Green, Blue)
   EnumLongestName(TypeInfo(TMyColor)) >
-------------------------------------------------------------------------------}
function TTypeUtils.EnumLongestName(pInfo: PTypeInfo; aOffset : Integer = 1): Integer;
var pData : PTypeData;
        i : Integer;
        l : Integer;
begin
  Result := 0;
  pData := GetTypeData(pInfo);

  if (pInfo^.Kind = tkSet)
    then pInfo := GetTypeData(pInfo)^.CompType^; //If a Set - get the enum type the set is based on

  if not (pInfo^.Kind = tkEnumeration)
    then DoRaise('TypeInfo not compatible', 'EnumLongestName');

  for i := 0 to pData^.MaxValue do
  begin
    l := Length(Copy(GetEnumName(pInfo,i), aOffset, $FF));
    if l>Result then Result := l;
  end;
end;

{-------------------------------------------------------------------------------
Returns a enumerated element as a string.
Example:
 type
   TMyColor = (Red, Green, Blue)
 var c : TMyColor;
 begin
   c := Green;
   EnumToString(TypeInfo(TMyColor), Integer(c)) > 'Green'
 end;
-------------------------------------------------------------------------------}
function TTypeUtils.EnumToString(pInfo: PTypeInfo; aOrdValue : Integer; aOffset : Integer = 1) : String;
var pData : PTypeData;
begin
  Result := '';
  pData := GetTypeData(pInfo);
  if aOffset<1 then aOffset := 1;
  if not (pInfo^.Kind = tkEnumeration)
    then DoRaise('TypeInfo not compatible', 'EnumToString');

  if aOrdValue>pData^.MaxValue then Exit;
  Result := Copy(GetEnumName(pInfo,aOrdValue), aOffset, $FFFF);
end;

{-------------------------------------------------------------------------------
Find the ordinal value of an enumerated type.
Example:
 type
   TMyColor = (clNone,clRed,clGreen,clBlue);
 var
   cl : TMyColor;
   Found : Boolean;
 begin
   cl:= TMyColor(StrToEnum(TypeInfo(TMyColor),'GREEN',Found)); Returns False and clNone
   cl:= TMyColor(StrToEnum(TypeInfo(TMyColor),'GREEN',Found,3)); Returns True amd clGreen
See also overload StrToEnum without EnumFound param
-------------------------------------------------------------------------------}
function TTypeUtils.StrToEnum(pInfo : PTypeInfo; const aElementName : String;out EnumFound : Boolean;aOffset : Integer; Ignorecase : Boolean) : Integer;
var pData : PTypeData;
        i : Integer;
begin
  Result := 0;
  EnumFound := False;
  pData := GetTypeData(pInfo);
  if aOffset<1 then aOffset := 1;
  if not (pInfo^.Kind = tkEnumeration)
    then DoRaise('TypeInfo not compatible', 'StrToEnum');

  for i := 0 to pData^.MaxValue do
  begin
    if IsSame(aElementName,Copy(GetEnumName(pInfo, i), aOffset, $FF), IgnoreCase) then
    begin
      Result := i;
      EnumFound := True;
      Break;
    end;
  end;
end;

{-------------------------------------------------------------------------------
Overloaded version without Out paramter.
Finds the ordinal value of an enumerated type by string while ignoring case.
Example:
 type
   TMyColor = (clNone, clRed, clGreen, clBlue);
 var
   cl : TMyColor;
 begin
   cl:= TMyColor(StrToEnumIC(TypeInfo(TMyColor),'clgreen'))
     or
   cl:= TMyColor(StrToEnumIC(TypeInfo(TMyColor),'green',3))
-------------------------------------------------------------------------------}
function TTypeUtils.StrToEnum(pInfo : PTypeInfo; const aElementName : String; aOffset : Integer; Ignorecase : Boolean) : Integer;
var Found : Boolean;
begin
  Result := StrToEnum(pInfo, aElementName, Found, aOffset, True);
end;

{-------------------------------------------------------------------------------
This function receives the TypInfo of a enumreated type or set, and returns the
size of the variable in bytes.
The size of Sets and Enums depends on the number of elements.
Enums are normally 1 byte (256 elements)
Sets can be upto 16 bytes.
  Example:
    EnumSize(TypeInfo(TComponentState)) >2
  "SizeOf(TComponentState)" will of course yield the same result, but in rtti
  routines you do not have the literal type.
-------------------------------------------------------------------------------}
function TTypeUtils.EnumSize(pInfo: PTypeInfo) : Integer;
var   p : PTypeInfo;
      n : Integer;
  pData : PTypeData;
begin
  Result := 0;
  if pInfo=nil then Exit;
  p := pInfo;
  if p^.Kind = tkSet
    then p := GetTypeData(p)^.CompType^;
  if not (p^.Kind = tkEnumeration) then Exit;
  pData := GetTypeData(p);
  n := pData^.MaxValue;
  if pInfo^.Kind = tkSet then
  begin
     Result := (n div 8)+1;
  end else
  begin
    Result := (n div 256)+1;
  end;
  if Result=3 then Result := 4;  //Delphi has 5,6,7,9..32 byte sets but never 3
end;


{-------------------------------------------------------------------------------
An enumerated type (especially a set) can be of different byte sizes.
This fuction supports a set with upto 64 elements (8 bytes).
The function is used by EnumSetToString to find the size of a set variable and
return the content as an Int64.
(A cast like Int64(Value) does not work on sets)
Example:
  TMyColor = (Red, Green, Blue);
  TMyColors = set of TMyColor;
var Colors : TMyColors = [Red, Blue];
  EnumValuesToInt(TypeInfo(TMyColors), @Colors)) > 5
-------------------------------------------------------------------------------}
function TTypeUtils.EnumValuesToInt(pInfo: PTypeInfo; aValue : Pointer) : Int64;
var aSize : Byte;
begin
  Result := 0;
  aSize := EnumSize(pInfo);
  if (aSize=0) or (aSize>SizeOf(Result)) then Exit;
  Move(aValue^, Result, aSize);
end;


{-------------------------------------------------------------------------------
 Returns the elements in a set as a String
 Example:
  type
    TMyColor = (Red, Green, Blue);
    TMyColors = set of TMyColor;
  var
    Colors : TMyColors;
  begin
    Colors := [Red, Blue];
    EnumSetToString(TypeInfo(TMyColors), @Colors) > 'Red,Blue'
  end;
-------------------------------------------------------------------------------}
function TTypeUtils.EnumSetToString(pInfo: PTypeInfo; aValue : Pointer; aOffset : Integer = 1; Brackets : Boolean = False; const aDelim : String = ',') : String;
begin
  Result := EnumSetToString(pInfo, EnumValuesToInt(pInfo, aValue), aOffset, Brackets, aDelim);
end;

{------------------------------------------------------------------------------
See overloaded function above.
-------------------------------------------------------------------------------}
function TTypeUtils.EnumSetToString(pInfo: PTypeInfo; aOrdValue : Int64; aOffset : Integer = 1; Brackets : Boolean = False; const aDelim : String = ',') : String;
var  i : Integer;
 pData : PTypeData;
begin
  Result := '';
  if (pInfo^.Kind <> tkSet)
    then DoRaise('TypeInfo is not a set', 'EnumSetToString');

  pInfo := GetTypeData(pInfo)^.CompType^; //now get the enum type the set is based on
  if pInfo^.Kind <> tkEnumeration
     then DoRaise('Enumeration expected', 'EnumSetToString');

  pData := GetTypeData(pInfo);
  if pData.MaxValue>63
    then DoRaise('Set is too large (more than 64 bits)', 'EnumSetToString'); //(More than 64 elements)
  if aOffset<1 then aOffset := 1;

  for i := 1 to pData.MaxValue+1 do 
  begin
    if Bit64On(aOrdValue, i) then
    begin
      if Result <> ''
        then Result := Result + aDelim;
      Result := Result + Copy(GetEnumName(pInfo, i-1), aOffset, $FFFF);
    end;
  end;
  if Brackets then
    Result := '[' + Result + ']';
end;


{-------------------------------------------------------------------------------
Returns the TypeData from the PropInfo record
-------------------------------------------------------------------------------}
function TTypeUtils.PropInfoToTypeData(aPropInfo : PPropInfo) : PTypeData;
begin
  Result := nil;
  if aPropInfo=nil then Exit;
  Result := GetTypeData(aPropInfo^.PropType^);
end;

{-------------------------------------------------------------------------------
Returns the TClass type if property is a tkClass
-------------------------------------------------------------------------------}
function TTypeUtils.PropInfoToClassType(aPropInfo : PPropInfo) : TClass;
var pData : PTypeData;
begin
  Result := nil;
  if aPropInfo=nil then Exit;
  if aPropInfo^.PropType^^.Kind<>tkClass then Exit;
  pData := PropInfoToTypeData(aPropInfo);
  if pData=nil then Exit;
  Result := pData^.ClassType;
end;

{-------------------------------------------------------------------------------
Returns true if ForClass is a decendant of the ClassType in PPropInfo
-------------------------------------------------------------------------------}
function TTypeUtils.ClassDecendantOf(aPropInfo : PPropInfo; ForClass : TClass) : Boolean;
var p : TClass;
begin
  Result := False;
  if ForClass=nil then Exit;
  p := PropInfoToClassType(aPropInfo);
  if p=nil then Exit;
  Result := p.InheritsFrom(ForClass);
end;


{-------------------------------------------------------------------------------
Example: PropInfoToPrimType(TypeInfo(TAnchors)) > ptSet
Note - not all types have TypeInfo - for example:
     Pointer, Class, Enumerated, Sets, Real48
If you declare a new type ie "CustNo = Integer" you also cannot compare TypeInfo(Integer)
This function therefore uses "two" passes to find the appropriate type.
-------------------------------------------------------------------------------}
function TTypeUtils.PropInfoToPrimType(aPropInfo : PPropInfo) : TPrimType;
  procedure FindDefault;
  var aBase : PTypeInfo;
  begin
    aBase := aPropInfo.PropType^;
    if aBase = TypeInfo(Byte) then Result := ptByte else
    if aBase = TypeInfo(ShortInt) then Result := ptShortInt else
    if aBase = TypeInfo(Boolean) then Result := ptBoolean else
    if aBase = TypeInfo(Char) then Result := ptChar else
    if aBase = TypeInfo(Word) then Result := ptWord else
    if aBase = TypeInfo(SmallInt) then Result := ptSmallInt else
    if aBase = TypeInfo(WideChar) then Result := ptWideChar else
    if aBase = TypeInfo(Integer) then Result := ptInteger else //ptLongint has Identical TypeInfo
    if aBase = TypeInfo(Cardinal) then Result := ptCardinal else //ptLongword has Identical TypeInfo
    if aBase = TypeInfo(Single) then Result := ptSingle else
    if aBase = TypeInfo(Real) then Result := ptReal else
    if aBase = TypeInfo(Int64) then Result := ptInt64 else
    if aBase = TypeInfo(Double) then Result := ptDouble else
    if aBase = TypeInfo(Comp) then Result := ptComp else
    if aBase = TypeInfo(Currency) then Result := ptCurrency else
    if aBase = TypeInfo(TDateTime) then Result := ptTDateTime else
    if aBase = TypeInfo(TDate) then Result := ptTDate else
    if aBase = TypeInfo(Extended) then Result := ptExtended else
    if aBase = TypeInfo(AnsiString) then Result := ptAnsiString else
    if aBase = TypeInfo(Variant) then Result := ptVariant else
    Result := ptNone;
  end;

  procedure GetOrdinal;
  var data : PTypeData;
  begin
    data := GetTypeData(aPropInfo.PropType^);
    case data^.OrdType of
      otSByte : Result := ptShortInt;
      otUByte : Result := ptByte;
      otSWord : Result := ptSmallInt;
      otUWord : Result := ptWord;
      otSLong : Result := ptInteger;
      otULong : Result := ptCardinal;
    end;
  end;

  procedure GetFloat;
  var data : PTypeData;
  begin
    data := GetTypeData(aPropInfo.PropType^);
    case data^.FloatType of
      ftSingle   : Result := ptSingle;
      ftDouble   : Result := ptDouble;
      ftExtended : Result := ptExtended;
      ftComp     : Result := ptComp;
      ftCurr     : Result := ptCurrency;
    end;
  end;

begin
  FindDefault;
  if Result <> ptNone then Exit;
  case aPropInfo.PropType^^.Kind of
    tkEnumeration :  Result := ptEnum;
    tkClass       : Result := ptClass;
    tkSet         : Result := ptSet;
    tkString      : Result := ptShortString; //AnsiString is tkLString (TypeInfo above works)
    tkLString     : Result := ptAnsiString;
    tkFloat       : GetFloat;
    tkChar        : Result := ptChar;
    tkWChar       : Result := ptWideChar;
    tkInteger     : GetOrdinal;
    tkVariant     : Result := ptVariant;
    tkInt64       : Result := ptInt64;
  end;
end;


{------------------------------------------------------------------------------
Returns the size of a primitive type.
Note for enumerated types and sets (which vary in size) use EnumSize.
-------------------------------------------------------------------------------}
function TTypeUtils.PrimTypeByteSize(aPrimType: TPrimType): Byte;
begin
  Result := 0;
  case aPrimType of
    ptNone          : Result := 0;
    ptByte          : Result := SizeOf(Byte);
    ptShortInt      : Result := SizeOf(ShortInt);
    ptBoolean       : Result := SizeOf(Boolean);
    ptChar          : Result := SizeOf(Char);
    ptWord          : Result := SizeOf(Word);
    ptSmallInt      : Result := SizeOf(SmallInt);
    ptWideChar      : Result := SizeOf(WideChar);
    ptInteger       : Result := SizeOf(Integer);
    ptLongint       : Result := SizeOf(Longint);
    ptLongword      : Result := SizeOf(Longword);
    ptCardinal      : Result := SizeOf(Cardinal);
    ptPointer       : Result := SizeOf(Pointer);   //pointer
    ptClass         : Result := SizeOf(TObject);   //pointer
    ptSingle        : Result := SizeOf(Single);
    ptInt64         : Result := SizeOf(Int64);
    ptDouble        : Result := SizeOf(Double);
    ptComp          : Result := SizeOf(Comp);
    ptCurrency      : Result := SizeOf(Currency);
    ptTDateTime     : Result := SizeOf(TDateTime);
    ptTDate         : Result := SizeOf(TDate);
    ptExtended      : Result := SizeOf(Extended);
    ptAnsiString    : Result := SizeOf(AnsiString); //pointer
    ptWideString    : Result := SizeOf(WideString); //pointer
    ptVariant       : Result := SizeOf(Variant);
    ptEnum          : Result := 0;  //depends  1..4 (Normally 1)
    ptSet           : Result := 0;  //depends  1..16
    ptShortString   : Result := 0;  //depends  1.256
    else DoRaise('Missing case','PrimTypeByteSize');
  end;
end;

{------------------------------------------------------------------------------
Use SetProperties to assign a value to a component and all child components.
You can optionally specify the Type of the property and the class holder type.
Example1:
  SetProperties(MyForm, 'Color', clYellow)
  This will set all components on the form to yellow - but you are assuming
  all properties named 'Color' are of type TMyColor.
Example2:
  SetProperties(MyForm, 'Color', clYellow, TypeInfo(TMyColor))
  This will set all the TMyColor properties of the components on the form to yellow.
Example3:
  SetProperties(MyForm, 'Color', clYellow, TypeInfo(TMyColor), TPanel)
  This will set all TPanel components on the form to yellow.
Example4:
  SetProperties(MyForm, 'Color', clYellow, TypeInfo(TMyColor), TPanel, True)
  This will set all TPanel based/decendant components on the form to yellow.
The function returns the number of properties/components assigned.
-------------------------------------------------------------------------------}
function TTypeUtils.SetProperties(aOwner : TComponent; const aPropName : string; const aValue : Variant; const OfType : PTypeInfo = nil; InClassType : TClass = nil; DoInherit : Boolean = False): Integer;
var PropList : TQPropList;

  procedure SetIt(aComponent : TComponent);
  var i : Integer;
  begin
    if InClassType<>nil then
    begin
      if DoInherit then
      begin
        if not (aComponent.InheritsFrom(InClassType)) then Exit;
      end else
      begin
        if InClassType<>aComponent.ClassType then Exit;
      end;
    end;
    PropList.ForInstance := aComponent;
    i := PropList.IndexByName(aPropName, OfType);
    if i<0 then Exit;
    PropList.Items[i].AsVariant := aValue;
    inc(Result);
  end;
var i : Integer;
begin
  Result := 0;
  PropList := TQPropList.Create;
  try
    for i := 0 to aOwner.ComponentCount-1
      do SetIt(aOwner.Components[i]);
  finally
    PropList.Free;
  end;
end;

{---------------------------------------------------------------------------------------------------
See notes at overloaded version above
---------------------------------------------------------------------------------------------------}
function TTypeUtils.SetProperties(aOwner : TComponent; const aPropName : string; const aValue : Variant; InClassType : TClass; DoInherit : Boolean = False): Integer;
begin
  Result := SetProperties(aOwner, aPropName, aValue, nil, InClassType, DoInherit);
end;

{ TQCustomPropItem }
constructor TQCustomPropItem.Create(ForAObject : PObjectOfProp; aPropInfo : PPropInfo);
begin
  inherited Create;
  DoSetProperty(ForAObject, aPropInfo);
end;

procedure TQCustomPropItem.DoSetProperty(ForAObject : PObjectOfProp; aPropInfo : PPropInfo);
{
Note on DirtyWrite (Default is False)
  If the write clause is missing from the property declaration we can use the GetProc address,
  for writing, if it refers directly to the variable, as in this example:
    private
      FName : String;
    published
      property Name : String read FName;
   We will first make sure that the read specifier is not a method with the GetDirectPtr function.
}
begin
  ClearInfoValues;
  try
    if ForAObject=nil then DoRaise('PropObject=nil');
    if ForAObject.FType=nil then DoRaise('ClassType=nil');
    if aPropInfo=nil then DoRaise('PropInfo=nil');
    FObjectPtr := ForAObject;
    FPropInfo := aPropInfo^;
    if not (FPropInfo.PropType^^.Kind in TypeManSupportedTypeKinds)
      then DoRaise('TypeKind not supported '+TypeUtils.EnumToString(TypeInfo(TTypeKind),Word(FPropInfo.PropType^^.Kind)));
    FTypeData := GetTypeData(FPropInfo.PropType^);
    FPrimType := DeterminePrimitiveType;
    if FPrimType=ptNone then DoRaise('Cannot determinine primpitive type for '+FPropInfo.Name);
    FTypeAccess := DetermineTypeAccess;
    FVarByteSize := DetermineVarByteSize;
    if GetDirtyWrite
      then SetDirtyWrite;
  except
    ClearInfoValues;
    DoRaise('', 'DoSetProperty');
  end;
end;

procedure TQCustomPropItem.ClearInfoValues;
begin
  FObjectPtr := nil; //external pointer
  FillChar(FPropInfo, SizeOf(FPropInfo),0);
  FTypeData := nil;
  FPrimType := ptNone;
  FTypeAccess := taNone;
end;


function TQCustomPropItem.GetDefaultValue : Integer;
begin
  if HasDefaultValue
    then Result := FPropInfo.Default
    else Result := 0;
end;

function TQCustomPropItem.GetHasDefaultValue : Boolean;
begin
  Result := (TypeKind in [tkInteger, tkChar, tkSet, tkEnumeration]);
  if Result then Result := FPropInfo.Default and $80000000=0;
end;

function TQCustomPropItem.DetermineTypeAccess : TQPropTypeAccess;
begin
  Result := taNone;
  case FPrimType of
    ptAnsiString, ptShortString :
      Result := taString;
    ptByte, ptShortInt, ptWord, ptSmallInt, ptInteger, ptLongint, ptLongword, ptCardinal :
      Result := taInt;
    ptInt64 :
      Result := taInt64;
    ptChar :
      Result := taChar;
    ptBoolean :
      Result := taBoolean;
    ptEnum :
      Result := taEnum;
    ptSet :
      Result := taSet;
    ptTDateTime, ptTDate :
      Result := taDateTime;
    ptSingle, ptDouble, ptComp, ptCurrency, ptExtended :
      Result := taFloat;
    ptVariant :
      Result := taVariant;
    ptClass :
      Result := taClass;
    else DoRaise('Missing case');
  end;
end;

function TQCustomPropItem.DeterminePrimitiveType : TPrimType;
begin
  Result := TypeUtils.PropInfoToPrimType(@FPropInfo);
end;

function TQCustomPropItem.DetermineVarByteSize: Byte;
begin
  Result := 0;
  if not DataOk then Exit;
  case FPrimType of
    ptEnum, ptSet : Result := TypeUtils.EnumSize(FPropInfo.PropType^);
    ptShortString : Result := FTypeData^.MaxLength+1;
    else Result := TypeUtils.PrimTypeByteSize(FPrimType);
  end;
end;

function TQCustomPropItem.DetermineDirectPtr : Pointer;
{
GetProc can either refer to the address of a method to get the propery value,
or to the offset of the field (variable) in the object's instance.
If the highest byte is $FF then we know it is not a method.
"property aProp read FValue" vs "property aProp read GetValue"
}
type aCast = record
  case byte of
    0 : (Bytes : array[0..3] of Byte);
    1 : (i : Cardinal);
    2 : (p : Pointer);
  end;
var c : aCast;
begin
  Result := nil;
  if ForInstance=nil then Exit;
  c.p := PropInfo.GetProc;
  if c.Bytes[3]<>$FF then Exit;
  c.i := c.i and $00FFFFFF;
  Inc(c.i, Cardinal(ForInstance));
  Result := c.p;
end;

function TQCustomPropItem.DataOk : Boolean;
begin
  Result := FPrimType<>ptNone;
end;

procedure TQCustomPropItem.SetDirtyWrite;
begin
  if DirectPtr=nil then Exit;
  if FPropInfo.SetProc<>nil then Exit;
  FPropInfo.SetProc := FPropInfo.GetProc
end;

procedure TQCustomPropItem.CheckWriteOk;
begin
  if FPrimType=ptNone then DoRaise('No TypeInfo');
  if FPropInfo.SetProc=nil then
  begin
    if (GetDirtyWrite) then SetDirtyWrite; //needed if ClassType is set in constructor, and "ForInstance" is set later
    if FPropInfo.SetProc=nil then DoRaise('Cannot write to property (ReadOnly?)');
  end;
  if ForInstance=nil then DoRaise('Class Instance not set');
end;

procedure TQCustomPropItem.CheckReadOk;
begin
  if FPrimType=ptNone then DoRaise('No TypeInfo');
  if FPropInfo.GetProc=nil
    then DoRaise('GetProc=nil');
  if ForInstance=nil then DoRaise('Class Instance not set');
end;

function TQCustomPropItem.GetName: String;
begin
  if DataOk
    then Result := FPropInfo.Name
    else Result := '';
end;

function TQCustomPropItem.GetTypeKind: TTypeKind;
begin
  if DataOk
    then Result := FPropInfo.PropType^^.Kind
    else Result := tkUnknown;
end;

function TQCustomPropItem.GetPrimTypeAsStr: String;
begin
  if DataOk
    then Result := TypeUtils.EnumToString(TypeInfo(TPrimType), Integer(FPrimType) ,3)
    else Result := '';
end;

function TQCustomPropItem.GetTypeName: String;
begin
  if DataOk
    then Result := FPropInfo.PropType^^.Name
    else Result := '';
end;


procedure TQCustomPropItem.DoRaise(const aMsg, aFunc: String);
begin
  RaiseWithPrevMsg(aMsg+#13+'Name: '+Name+#13+'Type: '+TypeName, aFunc, Self);
end;

function TQCustomPropItem.GetTypeInfoPtr: PTypeInfo;
begin
  if DataOk
    then Result := FPropInfo.PropType^
    else Result := nil;
end;

function TQCustomPropItem.GetAsVariant: Variant;
begin
  Result := Null;
  try
    CheckReadOk;
    case FTypeAccess of
      taVariant :
        Result := GetVariantProp(ForInstance, @FPropInfo);
      taString :
        Result := GetAsString;
      taInt :
        Result := GetAsInteger;
      taInt64 :
        Result := GetAsInt64+0.0;
      taChar :
        Result := GetAsChar;
      taBoolean :
        Result := GetAsBoolean;
      taEnum :
        Result := GetAsString;
      taSet :
        Result := GetAsString;
      taDateTime :
        Result := GetAsDateTime;
      taFloat :
        Result := GetAsFloat;
      else DoRaise(TypeName+' Not supported');
    end;
  except
    DoRaise('','GetAsVariant');
  end;
end;

procedure TQCustomPropItem.SetAsVariant(const Value: Variant);
begin
  try
    CheckWriteOk;
    case FTypeAccess of
      taVariant :
        SetVariantProp(ForInstance, @FPropInfo, Value);
      taString :
        SetAsString(Value);
      taInt :
        SetAsInteger(Value);
      taInt64 :
        SetAsInt64(Trunc(Value));
      taChar :
         SetAsString(Value);
      taBoolean :
        SetAsBoolean(Value);
      taEnum :
        SetAsString(Value);
      taSet :
        SetAsString(Value);
      taDateTime :
        SetAsDateTime(Value);
      taFloat :
        SetAsFloat(Value);
      else DoRaise(TypeName+' Not supported');
    end;
  except
    DoRaise('','SetAsVariant');
  end;
end;

function TQCustomPropItem.GetAsString: String;
begin
  Result := '';  //If Result := 'a' then GetSetProp will just append the result ??!!
  try
    CheckReadOk;
    case FTypeAccess of
      taString :
        Result := GetStrProp(ForInstance, @FPropInfo);
      taInt :
        if FPrimType in [ptLongword, ptCardinal]
          then Result := IntToStr(GetAsCardinal)
          else Result := IntToStr(GetAsInteger);
      taInt64 :
        Result := IntToStr(GetAsInt64);
      taChar :
        Result := GetAsChar;
      taBoolean : Result := BooleanIdents[AsBoolean];
      taEnum :
        Result := GetEnumProp(ForInstance, @FPropInfo);
      taSet :
        Result := GetSetProp(ForInstance, @FPropInfo);
      taDateTime :
        case FPrimType of
          ptTDate : Result := DateToStr(GetAsFloat)
               else Result := DateTimeToStr(GetAsFloat);
        end;
      taFloat :
        Result := FloatToStr(GetAsFloat);
      taVariant :
        Result := GetAsVariant;
      taClass :
        Result := Format('%p', [GetAsPointer]);
      else DoRaise(TypeName+' Not supported');
    end;
  except
    DoRaise('','GetAsString');
  end;
end;

procedure TQCustomPropItem.SetAsString(const Value: String);
begin
  try
    CheckWriteOk;
    case FTypeAccess of
      taString :
        SetStrProp(ForInstance, @FPropInfo, Value);
      taInt :
        SetAsInteger(StrToInt(Value));
      taInt64 :
        SetAsInt64(StrToInt64(Value));
      taChar :
        if Value=''
          then SetAsChar(#0)
          else SetAsChar(Value[1]);
      taBoolean :
        SetAsInteger(Byte(ToBool(Value)));
      taEnum :
        SetAsInteger(TypeUtils.StrToEnum(TypeInfoPtr, Value));
      taSet :
        SetSetProp(ForInstance, @FPropInfo, Value);
      taFloat :
        SetAsFloat(StrToFloat(Value));
      taDateTime :
        SetAsDateTime(VarToDateTime(Value));
      taVariant :
        SetAsVariant(Value);
      else DoRaise(TypeName+' Not supported');
    end;
  except
    DoRaise('','SetAsString');
  end;
end;

function TQCustomPropItem.GetAsInteger: Integer;
begin
  Result := 0;
  try
    CheckReadOk;
    case FTypeAccess of
      taInt :
        Result := GetOrdProp(ForInstance, @FPropInfo);
      taEnum :
        Result := GetOrdProp(ForInstance, @FPropInfo);
      taSet :
        Result := GetOrdProp(ForInstance, @FPropInfo);
      taChar :
        Result := GetOrdProp(ForInstance, @FPropInfo);
      taBoolean :
        Result := GetOrdProp(ForInstance, @FPropInfo);
      taInt64 :
        Result := GetAsInt64;
      taFloat :
        Result := Trunc(GetAsFloat);
      taString :
        Result := StrToInt(GetAsString);
      taVariant :
        Result := GetAsVariant;
      else DoRaise(TypeName+' Not supported');
    end;
  except
    DoRaise('','GetAsInteger');
  end;
end;

procedure TQCustomPropItem.SetAsInteger(const Value: Integer);
begin
  try
    CheckWriteOk;
    case FTypeAccess of
      taInt :
        SetOrdProp(ForInstance, @FPropInfo, Value);
      taEnum :
        SetOrdProp(ForInstance, @FPropInfo, Value);
      taSet :
        SetOrdProp(ForInstance, @FPropInfo, Value);
      taChar :
        SetOrdProp(ForInstance, @FPropInfo, Byte(Value));
      taBoolean :
        SetOrdProp(ForInstance, @FPropInfo, Integer(Boolean(Value>0)));  //make sure we store >1 as 1
      taInt64 :
        SetAsInt64(Value);
      taFloat :
        SetAsFloat(Value);
      taString :
        SetAsString(IntToStr(Value));
      taVariant :
        SetAsVariant(Value);
      else DoRaise(TypeName+' Not supported');
    end;
  except
    DoRaise('','SetAsInteger');
  end;
end;

function TQCustomPropItem.GetAsCardinal : Cardinal;
begin
  Result := GetAsInteger;
end;

procedure TQCustomPropItem.SetAsCardinal(const Value: Cardinal);
begin
  SetAsInteger(Value);
end;

function TQCustomPropItem.GetAsInt64: Int64;
begin
  Result := 0;
  try
    CheckReadOk;
    case FTypeAccess of
      taInt64 :
        Result := GetInt64Prop(ForInstance, @FPropInfo);
      taFloat :
        Result := Round(GetFloatProp(ForInstance, @FPropInfo));
      else Result := GetAsInteger;
    end;
  except
    DoRaise('','GetAsInt64');
  end;
end;

procedure TQCustomPropItem.SetAsInt64(const Value: Int64);
begin
  try
    CheckWriteOk;
    case FTypeAccess of
      taInt64 :
        SetInt64Prop(ForInstance, @FPropInfo, Value);
      taFloat :
        SetFloatProp(ForInstance, @FPropInfo, Value);
      else SetAsInteger(Value);
    end;
  except
    DoRaise('','SetAsInt64');
  end;
end;

function TQCustomPropItem.GetAsFloat: Extended;
begin
  Result := 0;
  try
    CheckReadOk;
    case FTypeAccess of
      taFloat :
        Result := GetFloatProp(ForInstance, @FPropInfo);
      taDateTime :
        Result := GetFloatProp(ForInstance, @FPropInfo);
      taVariant :
        Result := GetAsVariant;
      else Result := GetAsInt64;
    end;
  except
    DoRaise('','GetAsFloat');
  end;
end;

procedure TQCustomPropItem.SetAsFloat(const Value: Extended);
begin
  try
    CheckWriteOk;
    case FTypeAccess of
      taFloat :
        SetFloatProp(ForInstance, @FPropInfo, Value);
      taDateTime :
        SetFloatProp(ForInstance, @FPropInfo, Value);
      taVariant :
        SetAsVariant(Value);
      else SetAsInt64(Round(Value));
    end;
  except
    DoRaise('','SetAsFloat');
  end;
end;

function TQCustomPropItem.GetAsPointer: Pointer;
begin
  Result := nil;
  try
    CheckReadOk;
    case FTypeAccess of
      taInt :
        if VarByteSize=SizeOf(Pointer)
         then Result := Pointer(GetOrdProp(ForInstance, @FPropInfo))
         else DoRaise('Invalid type conversion');
      taClass :
        Result := Pointer(GetOrdProp(ForInstance, @FPropInfo));
      else DoRaise('Invalid type conversion');
    end;
  except
    DoRaise('','GetAsPointer');
  end;
end;

procedure TQCustomPropItem.SetAsPointer(const Value: Pointer);
begin
  try
    CheckWriteOk;
    case FTypeAccess of
      taInt :
        if VarByteSize=SizeOf(Pointer)
          then SetOrdProp(ForInstance, @FPropInfo, Integer(Value))
          else DoRaise('Invalid type conversion');
      taClass :
        SetOrdProp(ForInstance, @FPropInfo, Integer(Value));
      else DoRaise('Invalid type conversion');
    end;
  except
    DoRaise('','SetAsPointer');
  end;
end;

function TQCustomPropItem.GetAsClass: TObject;
begin
  Result := GetAsPointer;
end;

procedure TQCustomPropItem.SetAsClass(const Value: TObject);
begin
  SetAsPointer(Value);
end;

function TQCustomPropItem.GetAsBoolean: Boolean;
begin
  Result := Boolean(GetAsInteger);
end;

procedure TQCustomPropItem.SetAsBoolean(const Value: Boolean);
begin
  SetAsInteger(Byte(Value));
end;

function TQCustomPropItem.GetAsChar: Char;
begin
  Result := Char(GetAsInteger);
end;

procedure TQCustomPropItem.SetAsChar(const Value: Char);
begin
  SetAsInteger(Byte(Value));
end;

function TQCustomPropItem.GetAsDateTime: TDateTime;
begin
  Result := GetAsFloat;
end;

procedure TQCustomPropItem.SetAsDateTime(const Value: TDateTime);
begin
  SetAsFloat(Value);
end;

procedure TQCustomPropItem.ReadFromStream(aStream: TStream);

    function StreamInt : Integer;
    begin
      Result := 0;
      aStream.ReadBuffer(Result, VarByteSize);
    end;

    function StreamDouble : Double;
    begin
      aStream.ReadBuffer(Result, SizeOf(Result));
    end;

    function StreamTDateTime : TDateTime;
    begin
      aStream.ReadBuffer(Result, SizeOf(Result));
    end;
    
    function StreamExtended : Extended;
    begin
      aStream.ReadBuffer(Result, SizeOf(Result));
    end;

    function StreamSingle : Single;
    begin
      aStream.ReadBuffer(Result, SizeOf(Result));
    end;

    function StreamInt64 : Int64;
    begin
      aStream.ReadBuffer(Result, SizeOf(Result));
    end;

    function StreamString : String;
    var aLen : Integer;
    begin
      Result := '';
      aStream.ReadBuffer(aLen, SizeOf(aLen));
      if aLen<1 then Exit;
      if aLen>$fffff then DoRaise(IntToStr(aLen)+'> string constraint '+IntToStr($fffff));
      SetLength(Result, aLen);
      aStream.ReadBuffer(Result[1], aLen);
    end;

begin
  try
    case FTypeAccess of
      taInt, taChar, taBoolean, taEnum, taSet :
        AsInteger := StreamInt;
      taFloat, taDateTime :
        case PrimType of
          ptSingle    : AsFloat := StreamSingle;
          ptDouble    : AsFloat := StreamTDateTime;
          ptTDateTime : AsFloat := StreamDouble;
          ptTDate     : AsFloat := StreamDouble;
          else          AsFloat := StreamExtended;
        end;
      taInt64 :
        AsInt64 := StreamInt64;
      taString :
        AsString := StreamString;
      else DoRaise('Type not implemented');
    end;
  except
    DoRaise('','ReadFromStream');
  end;
end;

procedure TQCustomPropItem.WriteToStream(aStream: TStream);
    procedure StreamInt(const Value : Integer);
    begin
      aStream.WriteBuffer(Value, VarByteSize);
    end;

    procedure StreamSingle(const Value : Single);
    begin
      aStream.WriteBuffer(Value, SizeOf(Value));
    end;

    procedure StreamDouble(const Value : Double);
    begin
      aStream.WriteBuffer(Value, SizeOf(Value));
    end;

    procedure StreamTDateTime(const Value : TDateTime);
    begin
      aStream.WriteBuffer(Value, SizeOf(Value));
    end;

    procedure StreamExtended(const Value : Extended);
    begin
      aStream.WriteBuffer(Value, SizeOf(Value));
    end;

    procedure StreamInt64(const Value : Int64);
    begin
      aStream.WriteBuffer(Value, SizeOf(Value));
    end;

    procedure StreamString(const Value: String);
    var aLen : Integer;
    begin
      aLen := Length(Value);
      aStream.WriteBuffer(aLen, SizeOf(aLen));
      aStream.WriteBuffer(Value[1], aLen);
    end;

begin
  try
    case FTypeAccess of
      taInt, taChar, taBoolean, taEnum, taSet  :
        StreamInt(AsInteger);
      taFloat, taDateTime :
        case PrimType of
          ptSingle    : StreamSingle(AsFloat);
          ptDouble    : StreamTDateTime(AsFloat);
          ptTDateTime : StreamDouble(AsFloat);
          ptTDate     : StreamDouble(AsFloat);
          else          StreamExtended(AsFloat);
        end;
      taInt64 :
        StreamInt64(AsInt64);
      taString :
        StreamString(AsString);
      else DoRaise('Type not implemented');
    end;
  except
    DoRaise('','WriteToStream');
  end;
end;

procedure TQCustomPropItem.AssignValue(aSource : TQCustomPropItem);
begin
  if aSource=nil then DoRaise('aSource=nil');
  if aSource.TypeAccess=TypeAccess then
  begin
    case FTypeAccess of
      taVariant :
        AsVariant := aSource.AsVariant;
      taString :
        AsString := aSource.AsString;
      taInt :
        AsInteger := aSource.AsInteger;
      taInt64 :
        AsInt64 := aSource.AsInt64;
      taChar :
        AsChar := aSource.AsChar;
      taBoolean :
        AsBoolean := aSource.AsBoolean;
      taEnum :
        AsInteger := aSource.AsInteger;
      taSet :
        AsInteger := aSource.AsInteger;
      taFloat :
        AsFloat := aSource.AsFloat;
      taDateTime :
        AsFloat := aSource.AsDateTime;
      taClass :
        AsInteger := aSource.AsInteger;
      else DoRaise(TypeName+' Not supported');
    end;
  end else
  begin
    AsVariant := aSource.AsVariant;
  end;
end;

procedure TQCustomPropItem.AssignDefaultValue;
begin
  if not HasDefaultValue
    then DoRaise('No default value');
  AsInteger := DefaultValue;
end;


procedure TQCustomPropItem.AssignFromTField(aSource: TField);
begin
  if aSource=nil then DoRaise('aSource=nil');
  case aSource.DataType of
    ftString :
      AsString := aSource.AsString;
    ftSmallInt, ftWord, ftInteger :
      AsInteger := aSource.AsInteger;
    ftBoolean :
      AsBoolean := aSource.AsBoolean;
    ftFloat, ftCurrency :
      AsFloat := aSource.AsFloat;
    ftDateTime, ftDate :
      AsDateTime := aSource.AsDateTime;
    else DoRaise(TypeUtils.EnumToString(TypeInfo(TFieldType), Word(aSource.DataType))+ ' not supported');
  end;
end;

procedure TQCustomPropItem.AssignToTField(aTarget : TField);
begin
  if aTarget=nil then DoRaise('aTarget=nil');
  case aTarget.DataType of
    ftString :
      aTarget.AsString := AsString;
    ftSmallInt, ftWord, ftInteger :
      aTarget.AsInteger := AsInteger;
    ftBoolean :
      aTarget.AsBoolean := AsBoolean;
    ftFloat, ftCurrency :
      aTarget.AsFloat := AsFloat;
    ftDateTime, ftDate :
      aTarget.AsDateTime := AsDateTime;
    else DoRaise(TypeUtils.EnumToString(TypeInfo(TFieldType), Word(aTarget.DataType))+ ' not supported');
  end;
end;


procedure TQCustomPropItem.ClearValue;
begin
  try
    case FTypeAccess of
      taVariant :
        AsVariant := Null;
      taString :
        AsString := '';
      taInt :
        AsInteger := 0;
      taInt64 :
        AsInt64 := 0;
      taChar :
        AsChar := #0;
      taBoolean :
        AsBoolean := False;
      taEnum :
        AsInteger := 0;
      taSet :
        AsInteger := 0;
      taFloat :
        AsFloat := 0.0;
      taDateTime :
        AsFloat := 0.0;
      taClass :
        AsInteger := 0;
      else DoRaise(TypeName+' Not supported');
    end;
  except
    DoRaise('','SetAsVariant');
  end;
end;

function TQCustomPropItem.GetIsDirectAccess: Boolean;
begin
  Result := DirectPtr<>nil;
end;

function TQCustomPropItem.GetDirectPtr: Pointer;
begin
  Result := DetermineDirectPtr;
end;

function TQCustomPropItem.GetReadOnly: Boolean;
begin
  Result := FPropInfo.SetProc=nil;
end;

procedure TQCustomPropItem.SetForInstance(const Value: TObject);
begin
  if FObjectPtr=nil then DoRaise('FObjectOfProp=nil');
  if Value<>nil
    then if Value.ClassType<>ForType then DoRaise('ClassType mismatch');
  FObjectPtr.FInstance := Value;
end;

function TQCustomPropItem.GetForInstance: TObject;
begin
  if FObjectPtr=nil
    then Result := nil
    else Result := FObjectPtr.FInstance;
end;

function TQCustomPropItem.GetForType: TClass;
begin
  if FObjectPtr=nil
    then Result := nil
    else Result := FObjectPtr.FType;
end;

function TQCustomPropItem.GetDirtyWrite : Boolean;
begin
  if FObjectPtr=nil
    then Result := False
    else Result := FObjectPtr.FDirtyWrite;
end;

function TQCustomPropItem.GetPropInfoPtr: PPropInfo;
begin
  Result := @FPropInfo;
end;

function TQCustomPropItem.GetLongestEnumValue: Integer;
//if Property is of TMyColor = (Red, Green, Blue) (or set based on it) then the result will be 5 (length of 'Green')
begin
  if TypeAccess in [taEnum, taSet]
    then Result := TypeUtils.EnumLongestName(TypeInfoPtr)
    else Result := 0;
end;

function TQCustomPropItem.GetLongestSetValue: Integer;
//if Property is of set of TMyColor = (Red, Green, Blue) then the result will be 14 (length of 'Red,Green,Blue')
begin
  if TypeAccess=taSet
    then Result := Length(TypeUtils.EnumValuesToString(TypeInfoPtr, 1, ','))
    else Result := 0;
end;

function TQCustomPropItem.GetEnumValues : String;
//if Property is an enum (Red, Green, Blue) (or set based on it) then Result will be 'Red,Green,Blue'
begin
  if TypeAccess in [taEnum, taSet]
    then Result := TypeUtils.EnumValuesToString(TypeInfoPtr, 1, ',')
    else Result := '';
end;

function TQCustomPropItem.GetPropClassType: TClass;
begin
  if TypeAccess=taClass
    then Result := FTypeData.ClassType
    else Result := nil;
end;

function TQCustomPropItem.IsOfClass(aClass: TClass): Boolean;
begin
  if TypeAccess=taClass then
  begin
    if AsClass=nil
      then Result := FTypeData.ClassType.InheritsFrom(aClass)
      else Result := AsClass.InheritsFrom(aClass); //Solves problem of TStrings created as TStringList
  end else Result := False;
end;

function TQCustomPropItem.IsOfClass(aInstance : TObject): Boolean;
begin
  if aInstance=nil
    then Result := False
    else Result := IsOfClass(aInstance.ClassType);
end;

function TQCustomPropItem.GetHasMinMaxValue : Boolean;
begin
  if DataOk
   then Result := FPropInfo.PropType^^.Kind in [tkInteger, tkChar, tkEnumeration, tkWChar, tkInt64]
   else Result := False;
end;

function TQCustomPropItem.GetMinValue: Int64;
begin
  if HasMinMaxValue then
  begin
    if FPropInfo.PropType^^.Kind =tkInt64
      then Result := FTypeData^.MinInt64Value
      else Result := FTypeData^.MinValue
  end else Result := 0;
end;

function TQCustomPropItem.GetMaxValue: Int64;
begin
  if HasMinMaxValue then
  begin
    if FPropInfo.PropType^^.Kind =tkInt64
      then Result := FTypeData^.MaxInt64Value else
    begin
      if PrimType in [ptLongword, ptCardinal]    //MaxValue has a bug in Delphi 5,6 for Longwords and Cardinals
        then Result := $ffffffff
        else Result := FTypeData^.MaxValue
    end;
  end else Result := 0;
end;

function TQCustomPropItem.GetClassInUnit: String;
begin
  Result := '';
  if DataOk
    then if TypeKind=tkClass
      then Result := FTypeData^.UnitName;
end;

{ TQPropItem }
constructor TQPropItem.Create(ForAClassType: TClass; const aPropName: String; aDirtyWrite : Boolean = False);
begin
  inherited Create;
  SetProperty(ForAClassType, aPropName, aDirtyWrite);
end;


constructor TQPropItem.Create(ForAInstance: TObject; const aPropName: String; aDirtyWrite: Boolean);
begin
  inherited Create;
  SetProperty(ForAInstance, aPropName, aDirtyWrite);
end;


procedure TQPropItem.SetProperty(ForAClassType: TClass; const aPropName: String; aDirtyWrite : Boolean);
var p : PPropInfo;
begin
  FillChar(FObjectOfProp, SizeOf(FObjectOfProp), 0);

  if ForAClassType=nil
    then DoRaise('ForAClassType=nil');
  if ForAClassType.ClassInfo=nil
    then DoRaise(ForAClassType.ClassName+' has no ClassInfo, $M-');

  p := GetPropInfo(ForAClassType, aPropName);
  if p=nil then DoRaise('Cannot find property '+ForAClassType.ClassName+'.'+aPropName);

  FObjectOfProp.FType := ForAClassType;
  FObjectOfProp.FDirtyWrite := aDirtyWrite;
  DoSetProperty(@FObjectOfProp, p);
end;

procedure TQPropItem.SetProperty(ForAInstance: TObject; const aPropName: String; aDirtyWrite : Boolean);
var p : PPropInfo;
begin
  FillChar(FObjectOfProp, SizeOf(FObjectOfProp), 0);

  if ForAInstance=nil
    then DoRaise('ForAInstance=nil');
  if ForAInstance.ClassInfo=nil
    then DoRaise(ForAInstance.ClassName+' has no ClassInfo, $M-');

  p := GetPropInfo(ForAInstance, aPropName);
  if p=nil then DoRaise('Cannot find property '+ForAInstance.ClassName+'.'+aPropName);

  FObjectOfProp.FType := ForAInstance.ClassType;
  FObjectOfProp.FInstance := ForAInstance;
  FObjectOfProp.FDirtyWrite := aDirtyWrite;
  DoSetProperty(@FObjectOfProp, p);
end;

{ TQPropList }
constructor TQPropList.Create(aDirtyWrite : Boolean = False);
begin
  inherited Create;
  CreateVars;
  DirtyWrite := aDirtyWrite;
end;

constructor TQPropList.Create(ForAClassType : TClass; aDirtyWrite : Boolean = False; aTypeLimit : TPrimTypes = []; aClassLimit : TClass = nil);
begin
  inherited Create;
  CreateVars;
  FLimitTypes := aTypeLimit;
  FLimitClass := aClassLimit;
  DirtyWrite := aDirtyWrite;
  ForType := ForAClassType;
end;

constructor TQPropList.Create(ForAInstance : TObject; aDirtyWrite : Boolean = False; aTypeLimit : TPrimTypes = []; aClassLimit : TClass = nil);
begin
  inherited Create;
  CreateVars;
  FLimitTypes := aTypeLimit;
  FLimitClass := aClassLimit;
  DirtyWrite := aDirtyWrite;
  ForInstance := ForAInstance;
end;

constructor TQPropList.CreateExclude(ForAClassType : TClass; aDirtyWrite : Boolean = False; aTypeLimit : TPrimTypes = []; aClassLimit : TClass = nil);
begin
  inherited Create;
  CreateVars;
  FIsExcludeTypes := True;
  FLimitTypes := aTypeLimit;
  FLimitClass := aClassLimit;
  DirtyWrite := aDirtyWrite;
  ForType := ForAClassType;
end;

constructor TQPropList.CreateExclude(ForAInstance : TObject; aDirtyWrite : Boolean = False; aTypeLimit : TPrimTypes = []; aClassLimit : TClass = nil);
begin
  inherited Create;
  CreateVars;
  FIsExcludeTypes := True;
  FLimitTypes := aTypeLimit;
  FLimitClass := aClassLimit;
  DirtyWrite := aDirtyWrite;
  ForInstance := ForAInstance;
end;

destructor TQPropList.Destroy;
begin
  FreeVars;
  inherited Destroy;
end;

procedure TQPropList.CreateVars;
begin
  FList := TObjectList.Create;
end;

procedure TQPropList.FreeVars;
begin
  FList.Free;
end;

procedure TQPropList.AddPropertyToList(aPropInfo : PPropInfo);
{if the published property type(kind) is supported(usable) and not limited we add them to the list}
begin
  if not (aPropInfo^.PropType^^.Kind in TypeManSupportedTypeKinds) then Exit;
  if FLimitTypes<>[] then
  begin
    if (TypeUtils.PropInfoToPrimType(aPropInfo) in FLimitTypes)=FIsExcludeTypes then Exit;
  end;
  if FLimitClass<>nil then if aPropInfo^.PropType^^.Kind=tkClass then
  begin
    if TypeUtils.ClassDecendantOf(aPropInfo, FLimitClass)=FIsExcludeTypes then Exit;
  end;
  FList.Add(TQListPropItem.Create(@FObjectForProps, aPropInfo));
end;

procedure TQPropList.DoCreatePropertyList(ForAClassType : TClass; ForAInstance : TObject);
{
Loads published field properties to a TList of TQListPropItem.
Either ForAClassType or ForAInstance must be set.
}
var PropList : PPropList;
    Data     : PTypeData;
    i        : Integer;
begin
  Clear;
  if ForAInstance=nil then
  begin
    if ForAClassType=nil then DoRaise('ClassType=nil');
  end else
  begin
    if ForAClassType=nil
      then ForAClassType := ForAInstance.ClassType
      else if ForAClassType<>ForAInstance.ClassType then DoRaise('ClassType mismatch');
  end;
  if ForAClassType.ClassInfo=nil
    then DoRaise(ForAClassType.ClassName+' has no ClassInfo, $M-');

  FObjectForProps.FType := ForAClassType;
  FObjectForProps.FInstance := ForAInstance;

  Data := GetTypeData(ForType.ClassInfo);
  GetMem(PropList, Data^.PropCount*Sizeof(PPropInfo));
  try
    GetPropInfos(ForType.ClassInfo, PropList);
    for i := 0 to Data^.PropCount-1 do AddPropertyToList(PropList^[i]);
  finally
    FreeMem(PropList, Data^.PropCount*SizeOf(PPropInfo));
  end;
end;

procedure TQPropList.Clear;
var WasDirtyWrite : Boolean;
begin
  FList.Clear;
  WasDirtyWrite := FObjectForProps.FDirtyWrite;
  FillChar(FObjectForProps, SizeOf(FObjectForProps), 0);
  FObjectForProps.FDirtyWrite := WasDirtyWrite;
end;

function TQPropList.IndexByName(const aPropName: String; const MustBeOfType : PTypeInfo= nil): Integer;
var i : Integer;
begin
  Result := -1;
  for i := 0 to Flist.Count-1 do
  begin
    if AnsiCompareText(aPropName, TQPropItem(FList.Items[i]).Name) = 0 then
    begin
      if MustBeOfType<>nil
        then if Items[i].TypeInfoPtr<>MustBeOfType then Exit;
      Result := i;
      Exit;
    end;
  end;
end;

function TQPropList.PropExists(const aPropName: String; const MustBeOfType : PTypeInfo) : Boolean;
begin
  Result := IndexByName(aPropName, MustBeOfType)>=0;
end;

function TQPropList.FindProp(const aPropName: String; const MustBeOfType : PTypeInfo= nil) : TQListPropItem;
var i : Integer;
begin
  i := IndexByName(aPropName, MustBeOfType);
  if i<0
    then Result := nil
    else Result := Items[i];
end;


function TQPropList.ByName(const aPropName: String): TQListPropItem;
var i : Integer;
begin
  Result := nil;
  i := IndexByName(aPropName);
  if i>=0
    then Result := Items[i]
    else DoRaise('Property "'+aPropName+'" not found', 'PropByName');
end;

function TQPropList.GetPropByName(const Index: String): TQListPropItem;
begin
  Result := ByName(Index);
end;

function TQPropList.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TQPropList.GetItem(Index: Integer): TQListPropItem;
begin
  Result := TQListPropItem(FList[Index]);
end;

function TQPropList.GetPropName(Index: Integer): String;
begin
  Result := Items[Index].Name;
end;

function TQPropList.GetAsString(Index: Integer): String;
begin
  Result := Items[Index].AsString;
end;

procedure TQPropList.TypesToSL(aList: TStrings;Ident : Byte = 0; MaxValueStrLen : Integer = $ffff);
var i : Integer;
var s : String;
begin
  Assert(aList<>nil,'aList=nil');
  for i := 0 to Count -1 do
  begin
     s :=  RepeatChar(' ',Ident) + Format('%s : %s (%s, %d)',
      [PropName[i], Items[i].TypeName, Items[i].PrimTypeAsStr, Items[i].VarByteSize]);
     if Items[i].HasMinMaxValue
       then s := s + Format('(Min: %d, Max %d)',[Items[i].MinValue, Items[i].MaxValue]);
     if Items[i].TypeKind=tkClass
       then s := s + ' Unit='+Items[i].ClassInUnit;
     aList.Add(s);
  end;
end;

procedure TQPropList.ValuesToSL(aList : TStrings; WithType : Boolean = False; Ident : Byte = 0; MaxValueStrLen : Integer = $ffff); //for debugging
var i : Integer;
var Padded : String;
         v : String;
begin
  Padded := RepeatChar(' ',Ident);
  Assert(aList<>nil,'aList=nil');
  for i := 0 to Count -1 do
  begin
    v := AsString[i];
    if Length(v)>MaxValueStrLen then
    begin
      v := Copy(Trim(v),1,MaxValueStrLen)+'...';
    end;
    if WithType then
    begin
      aList.Add(Padded+Format('%s : %s (%s, %d) = %s',
       [PropName[i], Items[i].TypeName, Items[i].PrimTypeAsStr, Items[i].VarByteSize, v]));
    end else
    begin
      aList.Add(Padded+Format('%s = %s',
       [PropName[i], v]));
    end;
  end;
end;

procedure TQPropList.ReadFromStream(aStream: TStream; IgnoreTypes : TQPropTypesAccess = [taClass]);
var i : Integer;
begin
  try
    for i := 0 to Count-1 do
    begin
      if not (Items[i].TypeAccess in IgnoreTypes)
        then Items[i].ReadFromStream(aStream);
    end;
  except
    DoRaise('','ReadFromStream');
  end;
end;

procedure TQPropList.WriteToStream(aStream: TStream; IgnoreTypes : TQPropTypesAccess = [taClass]);
var i : Integer;
begin
  try
    for i := 0 to Count-1 do
    begin
      if not (Items[i].TypeAccess in IgnoreTypes)
        then Items[i].WriteToStream(aStream);
    end;
  except
    DoRaise('','WriteToStream');
  end;
end;

procedure TQPropList.ClearPropValues(IgnoreTypes : TQPropTypesAccess = [taClass]);
var i : Integer;
begin
  for i := 0 to Count-1 do
  begin
    if not (Items[i].TypeAccess in IgnoreTypes)
     then if not Items[i].ReadOnly
       then Items[i].ClearValue;
  end;
end;

procedure TQPropList.AssignPropValues(aSourceClass: TObject; IgnoreTypes : TQPropTypesAccess = [taClass]);
var tl : TQPropList;
begin
  tl := TQPropList.Create(aSourceClass);
  try
    AssignPropValues(tl);
  finally
    tl.Free;
  end;
end;

procedure TQPropList.AssignPropValues(aPropListForSourceClass  : TQPropList; IgnoreTypes : TQPropTypesAccess = [taClass]);
  procedure DoIt(aProp : TQListPropItem);
  var i : Integer;
  begin
    if aProp.TypeAccess in IgnoreTypes then Exit;
    i := aPropListForSourceClass.IndexByName(aProp.Name);
    if i<0 then Exit;
    if aPropListForSourceClass.Items[i].ReadOnly then Exit;
    aProp.AssignValue(aPropListForSourceClass.Items[i]);
  end;
var i : Integer;
begin
  for i := 0 to Count-1 do
  begin
    DoIt(Items[i]);
  end;
end;

procedure TQPropList.SetForType(const Value: TClass);
begin
  if Value=nil then
  begin
    Clear;
    Exit;
  end;
  if ForType=Value then Exit;
  DoCreatePropertyList(Value, nil);
end;

procedure TQPropList.SetForInstance(const Value: TObject);
begin
  try
    if Value=nil then
    begin
      FObjectForProps.FInstance := nil;
      Exit;
    end;
    if Value.ClassType <> ForType
      then Clear;
    if FList.Count=0
      then DoCreatePropertyList(nil, Value)
      else FObjectForProps.FInstance := Value;
  except
    DoRaise('',' SetForInstance');
  end;
end;

procedure TQPropList.AssignPropFromTField(aField : TField; RaiseIfPropMissing: Boolean);
var i : Integer;
begin
  i := IndexByName(aField.FieldName);
  if i<0 then
  begin
    if not RaiseIfPropMissing then Exit;
    raise exception.Create(aField.FieldName+' not in PropList');
  end;
  Items[i].AssignFromTField(aField);
end;

procedure TQPropList.AssignPropToDataSet(aProp : TQListPropItem; aDataSet : TDataSet; RaiseIfFieldMissing : Boolean);
var i : Integer;
    f : TField;
begin
  f := aDataSet.FindField(aProp.Name);
  if f=nil then
  begin
    if not RaiseIfFieldMissing then Exit;
    raise exception.Create(aProp.Name+' not in DataSet');
  end;
  aProp.AssignToTField(f)
end;

procedure TQPropList.FromDataSet(aDataSet: TDataSet; RaiseIfPropMissing: Boolean);
begin
  FromDataSet(aDataSet, [], False, RaiseIfPropMissing);
end;

procedure TQPropList.FromDataSet(aDataSet : TDataSet; const FieldNames : array of String; IsInclude : Boolean = True; RaiseIfPropMissing : Boolean = True);
{
 Assign fields in DataSet to properties with corresponding names.
 You can optionally specify include/exclude fields in the DataSet.
}
   function FieldInArray(const aName : String) : Boolean;
   var i : Integer;
   begin
     Result := True;
     for i := Low(FieldNames) to High(FieldNames) do
       if AnsiCompareText(aName, FieldNames[i]) = 0 then Exit;
     Result := False;
   end;
var i : Integer;
    f : TField;
begin
  if aDataSet=nil then DoRaise('aDataSet=nil');
  try
    if High(FieldNames)<0 then
    begin
      for i := 0 to aDataSet.FieldCount-1
        do AssignPropFromTField(aDataSet.Fields[i], RaiseIfPropMissing);
    end else
    if IsInclude then
    begin
      for i := Low(FieldNames) to High(FieldNames) do
      begin
        f := aDataSet.FindField(FieldNames[i]);
        if f=nil then DoRaise(FieldNames[i]+' not in DataSet');
        AssignPropFromTField(f, RaiseIfPropMissing);
      end;
    end else
    begin
      for i := 0 to aDataSet.FieldCount-1 do
      begin
        f := aDataSet.Fields[i];
        if not FieldInArray(f.FieldName)
          then AssignPropFromTField(f, RaiseIfPropMissing);
      end;
    end;
  except
    DoRaise(aDataSet.Name,'FromDataSet');
  end;
end;

procedure TQPropList.ToDataSet(aDataSet: TDataSet; RaiseIfFieldMissing: Boolean);
begin
  ToDataSet(aDataSet, [], False, RaiseIfFieldMissing);
end;

procedure TQPropList.ToDataSet(aDataSet : TDataSet; const PropNames : array of String; IsInclude : Boolean = True; RaiseIfFieldMissing : Boolean = True);
{
 Assign property fields to DataSet fields with corresponding names.
 You can optionally specify include/exclude fields in the PropList.
 Note DataSet must be in Edit mode.
}
   function FieldInArray(const aName : String) : Boolean;
   var i : Integer;
   begin
     Result := True;
     for i := Low(PropNames) to High(PropNames) do
       if AnsiCompareText(aName, PropNames[i]) = 0 then Exit;
     Result := False;
   end;
var i : Integer;
   pi : TQListPropItem;
begin
  if aDataSet=nil then DoRaise('aDataSet=nil');
  try
    if High(PropNames)<0 then
    begin
      for i := 0 to Count-1
        do AssignPropToDataSet(Items[i], aDataSet, RaiseIfFieldMissing);
    end else
    if IsInclude then
    begin
      for i := Low(PropNames) to High(PropNames) do
      begin
        pi := FindProp(PropNames[i]);
        if pi=nil then DoRaise(PropNames[i]+' not in PropList');
        AssignPropToDataSet(pi, aDataSet, RaiseIfFieldMissing);
      end;
    end else
    begin
      for i := 0 to Count-1 do
      begin
        pi := Items[i];
        if not FieldInArray(pi.Name)
          then AssignPropToDataSet(pi, aDataSet, RaiseIfFieldMissing);
      end;
    end;
  except
    DoRaise(aDataSet.Name,'FromDataSet');
  end;
end;

procedure TQPropList.AssignDefaultValues;
{If you declare an ordinal (or small set) property with the default clause; example:
  property MyColors : TMyColors read FMyColors write FMyColors default [red,green];
then this method will set all properties declared as "default" to their default values.}
var i : Integer;
begin
  if ForInstance=nil then DoRaise('ForInstance=nil');
  for i := 0 to Count-1 do
  begin
    if Items[i].HasDefaultValue
      then Items[i].AssignDefaultValue;
  end;
end;

procedure TQPropList.DoRaise(const aMsg, aFunc: String);
begin
  if ForType=nil
    then RaiseWithPrevMsg(aMsg, aFunc, Self)
    else RaiseWithPrevMsg(aMsg+#13#10+'ForType: '+ForType.ClassName, aFunc, Self)
end;


initialization
  TypeUtils := TTypeUtils.Create;

finalization
  TypeUtils.Free;
end.



